# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

import json
import logging
import re
from typing import Any, Dict, List, Optional, Tuple

import requests
import urllib3
import yaml

from .apic import ApicObject
from .const import RN_PREFIX_CLASSNAME_MAPPINGS
from .ndi import NDI

logger = logging.getLogger(__name__)


class PCV:
    def __init__(
        self,
        hostname_ip: str,
        username: str,
        password: str,
        domain: str,
        timeout: int,
    ):
        self.ndi = NDI(hostname_ip, username, password, domain, timeout)
        self.root = ApicObject("root", {}, [], None)
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def _resolve_tf_classnames(self, root: ApicObject, tf_plan: Any) -> None:
        """Helper function to resolve missing class names and key attributes using the Terraform plan"""
        if root.cl is None:
            dn = root["dn"]
            for change in tf_plan.get("resource_changes", []):
                section = (
                    "after" if change["change"].get("after") is not None else "before"
                )
                plan_dn = change["change"].get(section, {}).get("dn")
                if dn == plan_dn:
                    logger.debug(
                        "Resolving classname from Terraform plan for '{}'".format(dn)
                    )
                    root.cl = change["change"].get(section, {}).get("class_name")
                    name = (
                        change["change"].get(section, {}).get("content", {}).get("name")
                    )
                    if name:
                        logger.debug(
                            "Resolving name attribute from Terraform plan for '{}'".format(
                                dn
                            )
                        )
                        root.attributes["name"] = name

        for child in root.children:
            self._resolve_tf_classnames(child, tf_plan)

    def _resolve_static_classnames(self, root: ApicObject) -> None:
        """Helper function to resolve missing class names and key attributes using static mappings"""
        parts = str(root["dn"]).split("/")[-1].split("-", 1)
        prefix = parts[0]
        name = parts[1] if len(parts) > 1 else None
        if prefix in RN_PREFIX_CLASSNAME_MAPPINGS:
            mapping = RN_PREFIX_CLASSNAME_MAPPINGS[prefix]
            if root.cl is None:
                logger.debug(
                    "Statically resolving classname for '{}'".format(root["dn"])
                )
                root.cl = mapping.get("class")
            if root.cl == mapping.get("class"):
                for key in mapping.get("keys", []):
                    key_attribute = key.get("attribute")
                    key_regex = key.get("regex")
                    if (
                        key_attribute is not None
                        and key_regex is not None
                        and name is not None
                    ):
                        regex = re.compile(key_regex)
                        mo = regex.search(name)
                        if mo is not None and key_attribute not in root.attributes:
                            logger.debug(
                                "Statically adding key attribute '{}' for '{}'".format(
                                    key_attribute, root["dn"]
                                )
                            )
                            root.attributes[key_attribute] = mo.group()
        for child in root.children:
            self._resolve_static_classnames(child)

    def _check_classes(self, root: ApicObject) -> None:
        """Helper function to verify if all objects have classnames"""
        if root.cl is None:
            logger.error("Missing classname for '{}'".format(root["dn"]))
        for child in root.children:
            self._check_classes(child)

    def _load_json_objects(
        self, json_dict: Dict[Any, Any], parent: Optional[ApicObject] = None
    ) -> Optional[ApicObject]:
        """Helper function to load JSON objects into object tree"""
        new_obj = None
        for k, v in json_dict.items():
            new_obj = ApicObject(k, v.get("attributes"), [], parent)
            if parent:
                parent.children.append(new_obj)
            for child in v.get("children", []):
                self._load_json_objects(child, new_obj)
        return new_obj

    def load_json_files(self, filenames: List[str]) -> None:
        """Load objects from JSON files into object tree"""
        for filename in filenames:
            try:
                with open(filename, "r") as file:
                    inv = json.loads(file.read())
                    if "imdata" in inv:
                        for item in inv["imdata"]:
                            obj = self._load_json_objects(item)
                            self.root.insert(obj)
                    else:
                        obj = self._load_json_objects(inv)
                        self.root.insert(obj)
            except:  # noqa E722
                logger.error("Failed to load JSON file: {}".format(filename))
        self._resolve_static_classnames(self.root)
        self._check_classes(self.root)

    def load_tf_plan(self, filename: str) -> None:
        """Load changed objects from Terraform plan into object tree"""
        tf_plan = None
        with open(filename) as file:
            tf_plan = json.load(file)

        for change in tf_plan.get("resource_changes", []):
            if change.get("type") == "aci_rest_managed":
                action = change["change"].get("actions", [])
                if "create" in action or "update" in action or "delete" in action:
                    if "delete" in action:
                        classname = change["change"].get("before", {}).get("class_name")
                        attributes = {}
                        attributes = change["change"].get("before", {}).get("content")
                        attributes["status"] = "deleted"
                        attributes["dn"] = change["change"].get("before", {}).get("dn")
                    else:
                        classname = change["change"].get("after", {}).get("class_name")
                        attributes = change["change"].get("after", {}).get("content")
                        attributes["dn"] = change["change"].get("after", {}).get("dn")
                    attributes = {
                        k: v
                        for (k, v) in attributes.items()
                        if v != "" and v is not None
                    }
                    obj = ApicObject(classname, attributes, [], None)
                    self.root.insert(obj)

        self._resolve_static_classnames(self.root)
        self._resolve_tf_classnames(self.root, tf_plan)
        self._check_classes(self.root)

    def _write_pcv_events(self, events: List[Any], file: str) -> None:
        with open(file, "w") as fh:
            fh.write(yaml.dump(events, default_flow_style=False))

    def _write_pcv_url(self, url: str, file: str) -> None:
        with open(file, "w") as fh:
            fh.write(url)

    def ndi_pcv(
        self,
        name: str,
        group: str,
        site: str,
        suppress_events: str,
        file_summary: str,
        file_url: str,
    ) -> Tuple[Optional[requests.Response], Optional[List[Any]], Optional[str]]:
        """Trigger an NDI pre-change validation"""
        if not len(self.root.children):
            logger.info("No updates planned. No need to trigger a pre-change analysis.")
            return None, None, None
        logger.debug("Proposed change (JSON): {}".format(self.root[0]))
        err, job_id = self.ndi.start_pcv(name, group, site, str(self.root[0]))
        if err is not None:
            return err, None, None
        err, epoch_job_id = self.ndi.wait_pcv(group, site, str(job_id))
        if err is not None:
            return err, None, None
        err, events = self.ndi.get_pcv_results(
            group, site, str(epoch_job_id), suppress_events
        )
        if err is not None:
            return err, None, None
        err, url = self.ndi.get_pcv_url()
        if err is not None:
            return err, None, None
        if file_summary and events:
            self._write_pcv_events(events, file_summary)
        if file_url and url is not None:
            self._write_pcv_url(url, file_url)
        return None, events, url



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

from typing import Any, List, Mapping, Tuple

import click


class MutuallyExclusiveOption(click.Option):
    def __init__(self, *args: Any, **kwargs: Any):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs["help"] = help + (
                " NOTE: This argument is mutually exclusive with "
                " arguments: [" + ex_str + "]."
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(
        self, ctx: click.Context, opts: Mapping[str, Any], args: List[str]
    ) -> Tuple[Any, List[str]]:
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise click.UsageError(
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(self.name, ", ".join(self.mutually_exclusive))
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(ctx, opts, args)


hostname_ip = click.option(
    "-i",
    "--hostname-ip",
    type=str,
    envvar="PCV_HOSTNAME_IP",
    help="ND hostname or IP (required, env: PCV_HOSTNAME_IP).",
    required=True,
)

username = click.option(
    "-u",
    "--username",
    type=str,
    envvar="PCV_USERNAME",
    help="ND username (required, env: PCV_USERNAME).",
    required=True,
)

password = click.option(
    "-p",
    "--password",
    type=str,
    envvar="PCV_PASSWORD",
    help="ND password (required, env: PCV_PASSWORD).",
    prompt=True,
    hide_input=True,
    confirmation_prompt=True,
    required=True,
)

domain = click.option(
    "-d",
    "--domain",
    type=str,
    default="local",
    envvar="PCV_DOMAIN",
    help="ND login domain (optional, default: 'local', env: PCV_DOMAIN).",
    required=False,
)

group = click.option(
    "-g",
    "--group",
    type=str,
    default="default",
    envvar="PCV_GROUP",
    help="NDI insights group name (optional, default: 'default', env: PCV_GROUP).",
    required=False,
)

site = click.option(
    "-s",
    "--site",
    type=str,
    envvar="PCV_SITE",
    help="NDI site or fabric name (required, env: PCV_SITE).",
    required=True,
)

name = click.option(
    "-n",
    "--name",
    type=str,
    envvar="PCV_NAME",
    help="NDI pre-change validation name (required, env: PCV_NAME).",
    required=True,
)

timeout = click.option(
    "--timeout",
    type=int,
    default=15,
    envvar="PCV_TIMEOUT",
    help="NDI pre-change validation timeout in minutes (optional, default: 15, env: PCV_TIMEOUT).",
    required=False,
)

suppress_events = click.option(
    "--suppress-events",
    type=str,
    envvar="PCV_SUPPRESS_EVENTS",
    default="APP_EPG_NOT_DEPLOYED,APP_EPG_HAS_NO_CONTRACT_IN_ENFORCED_VRF",
    help="NDI comma-separated list of events to suppress (optional, default: 'APP_EPG_NOT_DEPLOYED,APP_EPG_HAS_NO_CONTRACT_IN_ENFORCED_VRF', env: PCV_SUPPRESS_EVENTS).",
    required=False,
)


file = click.option(
    "-f",
    "--file",
    cls=MutuallyExclusiveOption,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    multiple=True,
    envvar="PCV_FILE",
    help="NDI proposed change JSON file (optional, env: PCV_FILE).",
    required=False,
)

nac_tf_plan = click.option(
    "-t",
    "--nac-tf-plan",
    cls=MutuallyExclusiveOption,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    envvar="PCV_NAC_TF_PLAN",
    help="NDI proposed change Terraform plan output (optional, env: PCV_NAC_TF_PLAN).",
    required=False,
)

output_summary = click.option(
    "-o",
    "--output-summary",
    type=click.Path(exists=False, dir_okay=False, file_okay=True),
    envvar="PCV_OUTPUT_SUMMARY",
    help="NDI summary of new events/anomalies written to a file (optional, env: PCV_OUTPUT_SUMMARY).",
    required=False,
)

output_url = click.option(
    "-r",
    "--output-url",
    type=click.Path(exists=False, dir_okay=False, file_okay=True),
    envvar="PCV_OUTPUT_URL",
    help="NDI link (URL) to pre-change validation results written to a file (optional, env: PCV_OUTPUT_URL).",
    required=False,
)



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Daniel Schmidt <danischm@cisco.com>

from typing import Any, Dict

# Map of RN prefixes and its corresponding class name and key attributes
RN_PREFIX_CLASSNAME_MAPPINGS: Dict[str, Dict[str, Any]] = {
    "uni": {
        "class": "polUni",
    },
    "userext": {
        "class": "aaaUserEp",
    },
    "fabric": {
        "class": "fabricInst",
    },
    "hsPols": {
        "class": "healthPolCont",
    },
    "infra": {
        "class": "infraInfra",
    },
    "tn": {
        "class": "fvTenant",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "epg": {
        "class": "fvAEPg",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "rscons": {
        "class": "fvRsCons",
        "keys": [
            {
                "attribute": "tnVzBrCPName",
                "regex": ".*",
            }
        ],
    },
    "rsprov": {
        "class": "fvRsProv",
        "keys": [
            {
                "attribute": "tnVzBrCPName",
                "regex": ".*",
            }
        ],
    },
    "rsdomAtt": {
        "class": "fvRsDomAtt",
        "keys": [
            {
                "attribute": "tDn",
                "regex": r"(?<=\[).*(?=\])",
            }
        ],
    },
    "attenp": {
        "class": "infraAttEntityP",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "rsdomP": {
        "class": "infraRsDomP",
        "keys": [
            {
                "attribute": "tDn",
                "regex": r"(?<=\[).*(?=\])",
            }
        ],
    },
    "ap": {
        "class": "fvAp",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "BD": {
        "class": "fvBD",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "subnet": {
        "class": "fvSubnet",
        "keys": [
            {
                "attribute": "ip",
                "regex": r"(?<=\[).*(?=\])",
            }
        ],
    },
    "rsBDToOut": {
        "class": "fvRsBDToOut",
        "keys": [
            {
                "attribute": "tnL3extOutName",
                "regex": ".*",
            }
        ],
    },
    "brc": {
        "class": "vzBrCP",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "subj": {
        "class": "vzSubj",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "rssubjFiltAtt": {
        "class": "vzRsSubjFiltAtt",
        "keys": [
            {
                "attribute": "tnVzFilterName",
                "regex": ".*",
            }
        ],
    },
    "flt": {
        "class": "vzFilter",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "e": {
        "class": "vzEntry",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "out": {
        "class": "l3extOut",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "instP": {
        "class": "l3extInstP",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "extsubnet": {
        "class": "l3extSubnet",
        "keys": [
            {
                "attribute": "ip",
                "regex": r"(?<=\[).*(?=\])",
            }
        ],
    },
    "rttag": {
        "class": "l3extRouteTagPol",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "rspathAtt": {
        "class": "fvRsPathAtt",
        "keys": [
            {
                "attribute": "tDn",
                "regex": r"(?<=\[).*(?=\])",
            }
        ],
    },
    "leaves": {
        "class": "infraLeafS",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*(?=-typ-)",
            },
            {
                "attribute": "type",
                "regex": r"(?<=-typ-).*",
            },
        ],
    },
    "taboo": {
        "class": "vzTaboo",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "destgrp": {
        "class": "spanDestGrp",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "srcgrp": {
        "class": "spanSrcGrp",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "spanlbl": {
        "class": "spanSpanLbl",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "ctx": {
        "class": "fvCtx",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "pol": {
        "class": "coopPol",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
    "format": {
        "class": "datetimeFormat",
        "keys": [
            {
                "attribute": "name",
                "regex": ".*",
            }
        ],
    },
}
